# Author; alin m elena, alin@elena.re
# Contribs;
# Date: 16-11-2024
# ©alin m elena,
"""pack molecules inside various shapes."""

from __future__ import annotations

from pathlib import Path

from ase import Atoms
from ase.build import molecule as build_molecule
from ase.io import read, write
from ase.units import kB
from janus_core.calculations.geom_opt import GeomOpt
from janus_core.calculations.md import NVE
from janus_core.helpers.mlip_calculators import choose_calculator
from numpy import cos, exp, pi, random, sin, sqrt


def random_point_in_sphere(c: (float, float, float), r: float) -> (float, float, float):
    """
    Generate a random point inside a sphere of radius r, centered at c.

    Parameters
    ----------
        c (tuple): The center of the sphere as (x, y, z).
        r (float): The radius of the sphere.

    Returns
    -------
        tuple: A point (x, y, z) inside the sphere.
    """
    rad = r * random.rand() ** (1 / 3)

    theta = random.uniform(0, 2 * pi)
    phi = random.uniform(0, pi)

    x = c[0] + rad * sin(phi) * cos(theta)
    y = c[1] + rad * sin(phi) * sin(theta)
    z = c[2] + rad * cos(phi)

    return (x, y, z)


def random_point_in_ellipsoid(
    d: (float, float, float), a: float, b: float, c: float
) -> (float, float, float):
    """
    Generate a random point inside an ellipsoid with axes a, b, c, centered at d.

    Parameters
    ----------
        d (tuple): The center of the ellipsoid as (x, y, z).
        a (float): The semi-axis length of the ellipsoid along the x-axis.
        b (float): The semi-axis length of the ellipsoid along the y-axis.
        c (float): The semi-axis length of the ellipsoid along the z-axis.

    Returns
    -------
        tuple: A point (x, y, z) inside the ellipsoid.
    """
    theta = random.uniform(0, 2 * pi)
    phi = random.uniform(0, pi)
    rad = random.rand() ** (1 / 3)

    x = d[0] + a * rad * sin(phi) * cos(theta)
    y = d[1] + b * rad * sin(phi) * sin(theta)
    z = d[2] + c * rad * cos(phi)

    return (x, y, z)


def random_point_in_box(
    d: (float, float, float), a: float, b: float, c: float
) -> (float, float, float):
    """
    Generate a random point inside a box with sides a, b, c, centered at d.

    Parameters
    ----------
        d (tuple): The center of the box as (x, y, z).
        a (float): The length of the box along the x-axis.
        b (float): The length of the box along the y-axis.
        c (float): The length of the box along the z-axis.

    Returns
    -------
        tuple: A point (x, y, z) inside the box.
    """
    x = d[0] + random.uniform(-a * 0.5, a * 0.5)
    y = d[1] + random.uniform(-b * 0.5, b * 0.5)
    z = d[2] + random.uniform(-c * 0.5, c * 0.5)

    return (x, y, z)


def random_point_in_cylinder(
    c: (float, float, float), r: float, h: float, d: str
) -> (float, float, float):
    """
    Generate a random point inside a cylinder with radius r and height h, centered at c.

    Parameters
    ----------
        c (tuple): The center of the cylinder as (x, y, z).
        r (float): The radius of the cylinder's base.
        h (float): The height of the cylinder.
        direction (str): direction along which cylinger is oriented

    Returns
    -------
        tuple: A point (x, y, z) inside the cylinder.
    """
    theta = random.uniform(0, 2 * pi)
    rad = r * sqrt(random.rand())

    if d == "z":
        z = c[2] + random.uniform(-h * 0.5, h * 0.5)
        x = c[0] + rad * cos(theta)
        y = c[1] + rad * sin(theta)
    elif d == "y":
        y = c[1] + random.uniform(-h * 0.5, h * 0.5)
        x = c[0] + rad * cos(theta)
        z = c[2] + rad * sin(theta)
    elif d == "x":
        x = c[0] + random.uniform(-h * 0.5, h * 0.5)
        y = c[1] + rad * sin(theta)
        z = c[2] + rad * cos(theta)

    return (x, y, z)


def validate_value(label: str, x: float | int) -> None:
    """Validate input value, and raise an exception."""
    if x is not None and x < 0.0:
        err = f"Invalid {label}, needs to be positive"
        print(err)
        raise Exception(err)


def set_random_seed(seed: int) -> None:
    """Set random seed."""
    random.seed(seed)


def set_defaults(
    cell: (float, float, float),
    centre: (float, float, float) | None = None,
    where: str | None = None,
    a: float | None = None,
    b: float | None = None,
    c: float | None = None,
    radius: float | None = None,
    height: float | None = None,
) -> tuple(
    (float, float, float),
    float | None,
    float | None,
    float | None,
    float | None,
    float | None,
):
    """Set defaults for insertion areas."""
    if centre is None:
        centre = (cell[0] * 0.5, cell[1] * 0.5, cell[2] * 0.5)

    if where == "anywhere":
        a, b, c = cell[0], cell[1], cell[2]
    elif where == "sphere":
        radius = radius or min(cell) * 0.5
    elif where == "cylinderZ":
        radius = radius or min(cell[0], cell[1]) * 0.5
        height = height or 0.5 * cell[2]
    elif where == "cylinderY":
        radius = radius or min(cell[0], cell[2]) * 0.5
        height = height or 0.5 * cell[1]
    elif where == "cylinderX":
        radius = radius or min(cell[2], cell[1]) * 0.5
        height = height or 0.5 * cell[0]
    elif where == "box":
        a, b, c = a or cell[0], b or cell[1], c or cell[2]
    elif where == "ellipsoid":
        a, b, c = a or cell[0] * 0.5, b or cell[1] * 0.5, c or cell[2] * 0.5
    return (centre, a, b, c, radius, height)


def pack_molecules(
    system: str | Atoms = None,
    molecule: str = "H2O",
    nmols: int = -1,
    arch: str = "cpu",
    model: str = "mace_mp",
    device: str = "medium-omat-0",
    where: str = "anywhere",
    center: tuple[float, float, float] = None,
    radius: float = None,
    height: float = None,
    a: float = None,
    b: float = None,
    c: float = None,
    seed: int = 2025,
    temperature: float = 300.0,
    ntries: int = 50,
    geometry: bool = False,
    fmax: float = 0.1,
    cell_a: float = None,
    cell_b: float = None,
    cell_c: float = None,
    out_path: str = ".",
    every: int = -1,
    relax_strategy: str = "geometry_optimisation",
    insert_strategy: str = "random",
    md_steps: int = 10,
    md_timestep: float = 1.0,
    md_temperature: float = 100.0,
) -> tuple(float, Atoms):
    """
    Pack molecules into a system based on the specified parameters.

    Parameters
    ----------
        system (str|Atoms): Path to the system file or name of the system.
        molecule (str): Path to the molecule file or name of the molecule.
        nmols (int): Number of molecules to insert.
        arch (str): Architecture for the calculator.
        model (str): Path to the model file.
        device (str): Device to run calculations on (e.g., "cpu" or "cuda").
        where (str): Region to insert molecules ("anywhere",
                     "sphere", "cylinderZ", etc.).
        center (Optional[Tuple[float, float, float]]): Center of the insertion region.
        radius (Optional[float]): Radius for spherical or cylindrical insertion.
        height (Optional[float]): Height for cylindrical insertion.
        a, b, c (Optional[float]): Parameters for box or ellipsoid insertion.
        seed (int): Random seed for reproducibility.
        temperature (float): Temperature in Kelvin for acceptance probability.
        ntries (int): Maximum number of attempts to insert each molecule.
        geometry (bool): Whether to perform geometry optimization after insertion.
        cell_a, cell_b, cell_c (float): Cell dimensions if system is empty.
        out_path (str): path to save various outputs
        every (int): After how many instertions to do a relaxation,
                     default -1 means none..
        md_temperature (float): Temperature in Kelvin for MD.
        md_steps (int): Number of steps for MD.
        md_timestep (float): Timestep in fs for MD.
        insert_strategy (str): Insert strategy, "random" or "md"
        relax_strategy (str): Relax strategy, "geometry_optimisation" or "md"

    Returns
    -------
        tuple: A tuple energy and Atoms object containing
               original system and added molecules..

    """
    kbt = temperature * kB
    validate_value("temperature", temperature)
    validate_value("radius", radius)
    validate_value("height", height)
    validate_value("fmax", fmax)
    validate_value("seed", seed)
    validate_value("box a", a)
    validate_value("box b", b)
    validate_value("box c", c)
    validate_value("ntries", ntries)
    validate_value("cell box cell a", cell_a)
    validate_value("cell box cell b", cell_b)
    validate_value("cell box cell c", cell_c)
    validate_value("nmols", nmols)
    validate_value("MD steps", md_steps)
    validate_value("MD timestep", md_timestep)
    validate_value("MD temperature", md_temperature)

    set_random_seed(seed)

    if system is None:
        sys = Atoms(cell=[cell_a, cell_b, cell_c], pbc=[True, True, True])
        sysname = ""
    elif isinstance(system, Atoms):
        sys = system.copy()
        sysname = sys.get_chemical_formula()
    else:
        sys = read(system)
        sysname = Path(system).stem + "+"

    # Print summary
    print(f"Inserting {nmols} {molecule} molecules in {sysname}.")
    print(f"Using {arch} model {model} on {device}.")
    print(f"Insert in {where}.")

    cell = sys.cell.lengths()

    center, a, b, c, radius, height = set_defaults(
        cell, center, where, a, b, c, radius, height
    )

    calc = choose_calculator(arch=arch, model_path=model, device=device)
    sys.calc = calc

    e = sys.get_potential_energy() if len(sys) > 0 else 0.0

    csys = sys.copy()
    i = 0
    while i < nmols:
        accept = False
        for _itry in range(ntries):
            mol = load_molecule(molecule)
            tv = get_insertion_position(where, center, a, b, c, radius, height)
            mol = rotate_molecule(mol)
            mol.translate(tv)

            tsys = csys.copy() + mol.copy()
            if insert_strategy == "hmc":
                tsys = run_md_nve(
                    tsys, md_temperature, md_steps, md_timestep, arch, model, device
                )

            if every > 0 and _itry / every == 0:
                tsys = save_the_day(
                    struct_path=tsys,
                    device=device,
                    arch=arch,
                    model=model,
                    fmax=fmax,
                    out_path=out_path,
                    md_temperature=md_temperature,
                    md_steps=md_steps,
                    md_timestep=md_timestep,
                    relax_strategy=relax_strategy,
                )

            tsys.calc = calc
            en = tsys.get_potential_energy()
            de = en - e

            acc = exp(-de / kbt)
            u = random.random()
            print(f"Old energy={e}, new energy={en}, {de=}, {acc=}, random={u}")

            if u <= acc:
                accept = True
                break
        if accept:
            csys = tsys.copy()
            e = en
            i += 1
            print(f"Inserted particle {i}")
            write(Path(out_path) / f"{sysname}{i}{Path(molecule).stem}.cif", csys)
        else:
            # Things are bad, maybe geomatry optimisation saves us
            # once you hit here is bad, this can keep looping
            print(f"Failed to insert particle {i + 1} after {ntries} tries")
            csys = save_the_day(
                csys,
                device,
                arch,
                model,
                fmax,
                out_path,
                md_temperature,
                md_steps,
                md_timestep,
                relax_strategy,
            )

    energy_final = e

    # Perform final geometry optimization if requested
    if geometry:
        energy_final, csys = optimize_geometry(
            Path(out_path) / f"{sysname}{nmols}{Path(molecule).stem}.cif",
            device,
            arch,
            model,
            fmax,
            out_path,
            True,
        )
    return (energy_final, csys)


def load_molecule(molecule: str):
    """Load a molecule from a file or build it."""
    try:
        return build_molecule(molecule)
    except KeyError:
        return read(molecule)


def get_insertion_position(
    where: str,
    center: tuple[float, float, float],
    a: float = None,
    b: float = None,
    c: float = None,
    radius: float = None,
    height: float = None,
) -> tuple[float, float, float]:
    """Get a random insertion position based on the region."""
    if where == "sphere":
        return random_point_in_sphere(center, radius)
    if where == "box":
        return random_point_in_box(center, a, b, c)
    if where == "ellipsoid":
        return random_point_in_ellipsoid(center, a, b, c)
    if where in ["cylinderZ", "cylinderY", "cylinderX"]:
        axis = where[-1].lower()
        return random_point_in_cylinder(center, radius, height, axis)
    return random.random(3) * [a, b, c]


def rotate_molecule(mol):
    """Rotate a molecule randomly."""
    ang = random.random(3)
    mol.euler_rotate(
        phi=ang[0] * 360, theta=ang[1] * 180, psi=ang[2] * 360, center=(0.0, 0.0, 0.0)
    )
    return mol


def save_the_day(
    struct_path: str | Atoms,
    device: str = "",
    arch: str = "",
    model: str = "",
    fmax: float = 0.01,
    out_path: str = ".",
    md_temperature: float = 100.0,
    md_steps: int = 10,
    md_timestep: float = 1.0,
    relax_strategy: str = "geometry_optimisation",
) -> Atoms:
    """Geometry optimisation or MD to get a better structure."""
    if relax_strategy == "geometry_optimisation":
        _, a = optimize_geometry(
            struct_path,
            device,
            arch,
            model,
            fmax,
            out_path,
        )
        return a
    if relax_strategy == "md":
        return run_md_nve(
            struct_path, md_temperature, md_steps, md_timestep, arch, model, device
        )
    return None


def run_md_nve(
    struct_path: str | Atoms,
    temp: float = 100.0,
    steps: int = 10,
    timestep: float = 1.0,
    arch: str = "",
    model: str = "",
    device: str = "",
) -> Atoms:
    """Run nve simulation."""
    if isinstance(struct_path, Atoms):
        md = NVE(
            struct=struct_path,
            temp=temp,
            device=device,
            arch=arch,
            calc_kwargs={"model_paths": model},
            stats_every=1,
            steps=steps,
            timestep=timestep,
        )
    else:
        md = NVE(
            struct_path=struct_path,
            temp=temp,
            device=device,
            arch=arch,
            calc_kwargs={"model_paths": model},
            stats_every=1,
            steps=steps,
            timestep=timestep,
        )
    md.run()
    return md.struct


def optimize_geometry(
    struct_path: str | Atoms,
    device: str,
    arch: str,
    model: str,
    fmax: float,
    out_path: str = ".",
    opt_cell: bool = False,
) -> tuple(float, Atoms):
    """Optimize the geometry of a structure."""
    if isinstance(struct_path, Atoms):
        geo = GeomOpt(
            struct=struct_path,
            device=device,
            arch=arch,
            fmax=fmax,
            calc_kwargs={"model_paths": model},
            filter_kwargs={"hydrostatic_strain": opt_cell},
        )
        geo.run()
    else:
        geo = GeomOpt(
            struct_path=struct_path,
            device=device,
            arch=arch,
            fmax=fmax,
            calc_kwargs={"model_paths": model},
            filter_kwargs={"hydrostatic_strain": opt_cell},
        )
        geo.run()
        write(Path(out_path) / f"{Path(struct_path).stem}-opt.cif", geo.struct)
    return (geo.struct.get_potential_energy(), geo.struct)



"""Utility functions for tests."""

# lifted from janus_core github.com/stfc/janus-core/
from __future__ import annotations

import re


def strip_ansi_codes(output: str) -> str:
    """
    Remove any ANSI sequences from output string.

    Based on:
    https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python/14693789#14693789

    Parameters
    ----------
    output
        Output that may contain ANSI sequences to be removed.

    Returns
    -------
    str
        Output with ANSI sequences removed.
    """
    # 7-bit C1 ANSI sequences
    ansi_escape = re.compile(
        r"""
        \x1B  # ESC
        (?:   # 7-bit C1 Fe (except CSI)
            [@-Z\\-_]
        |     # or [ for CSI, followed by a control sequence
            \[
            [0-?]*  # Parameter bytes
            [ -/]*  # Intermediate bytes
            [@-~]   # Final byte
        )
    """,
        re.VERBOSE,
    )
    return ansi_escape.sub("", output)



"""Advanced tests for packmm."""

# Author; alin m elena, alin@elena.re
# Contribs;
# Date: 23-02-2025
# ©alin m elena,
from __future__ import annotations

from ase.build import bulk
from ase.io import read, write
import pytest
from typer.testing import CliRunner

from pack_mm.cli.packmm import app

runner = CliRunner()

err = 1.0e-8


def test_packmm_hmc(tmp_path):
    """Check values."""
    result = runner.invoke(
        app,
        [
            "--nmols",
            "2",
            "--model",
            "small-0b2",
            "--insert-strategy",
            "hmc",
            "--seed",
            "2042",
            "--cell-a",
            "10",
            "--cell-b",
            "10",
            "--cell-c",
            "10",
            "--out-path",
            tmp_path,
        ],
    )
    assert result.exit_code == 0
    assert (tmp_path / "1H2O.cif").exists()
    assert (tmp_path / "2H2O.cif").exists()
    f = read(tmp_path / "2H2O.cif")
    assert f[0].position == pytest.approx([7.83420049, 6.21661184, 5.21814887], abs=err)
    assert (tmp_path / "2H2O-opt.cif").exists()
    f = read(tmp_path / "2H2O-opt.cif")
    assert f[0].position == pytest.approx([7.83120709, 6.21742874, 5.22597213], abs=err)


def test_packmm_every(tmp_path):
    """Check values."""
    na = bulk("NaCl", "rocksalt", a=5.61, cubic=True)
    write(tmp_path / "system.cif", na)
    write("system.cif", na)

    result = runner.invoke(
        app,
        [
            "--system",
            tmp_path / "system.cif",
            "--molecule",
            "H2",
            "--every",
            "1",
            "--nmols",
            "2",
            "--model",
            "small-0b2",
            "--seed",
            "2042",
            "--out-path",
            tmp_path,
            "--no-geometry",
        ],
    )
    assert result.exit_code == 0
    assert (tmp_path / "system.cif").exists()
    assert (tmp_path / "system+1H2.cif").exists()
    assert (tmp_path / "system+2H2.cif").exists()
    f = read(tmp_path / "system+2H2.cif")
    assert f[0].position == pytest.approx([1.24701529, 0.024216, 0.11632905], abs=err)



"""Tests for core."""

# -*- coding: utf-8 -*-
# Author; alin m elena, alin@elena.re
# Contribs;
# Date: 22-02-2025
# ©alin m elena,
from __future__ import annotations

from ase import Atoms
from ase.build import molecule as build_molecule
from ase.io import write
from numpy import random
import pytest

from pack_mm.core.core import (
    get_insertion_position,
    load_molecule,
    optimize_geometry,
    pack_molecules,
    random_point_in_box,
    random_point_in_cylinder,
    random_point_in_ellipsoid,
    random_point_in_sphere,
    rotate_molecule,
    run_md_nve,
    save_the_day,
    set_defaults,
    validate_value,
)

err = 1.0e-8


# Set a fixed seed for reproducibility in tests
@pytest.fixture(autouse=True)
def set_random_seed():
    """Set random seed."""
    random.seed(2042)


def test_random_point_in_sphere():
    """Test point in sphere."""
    center = (2, 2, 2)
    radius = 8.0
    x, y, z = random_point_in_sphere(center, radius)
    assert x == pytest.approx(-3.299696236298196, abs=err)
    assert y == pytest.approx(-3.046619861327052, abs=err)
    assert z == pytest.approx(1.1884891239565165, abs=err)


def test_random_point_in_ellipsoid():
    """Test point in ellipsoid."""
    center = (0, 0, 0)
    a, b, c = 2.0, 2.0, 4.0
    x, y, z = random_point_in_ellipsoid(center, a, b, c)
    assert x == pytest.approx(0.27914659851849705, abs=err)
    assert y == pytest.approx(-1.4815802529721946, abs=err)
    assert z == pytest.approx(-1.2059956538672925, abs=err)


def test_random_point_in_box():
    """Test point in box."""
    center = (0, 0, 0)
    a, b, c = 1.0, 2.0, 3.0
    x, y, z = random_point_in_box(center, a, b, c)
    assert x == pytest.approx(0.2796391455704136, abs=err)
    assert y == pytest.approx(0.24221552377157196, abs=err)
    assert z == pytest.approx(0.10546160764197499, abs=err)


def test_random_point_in_cylinder():
    """Test point in cylinder."""
    center = (0, 0, 0)
    radius = 2.0
    height = 5.0

    direction = "z"
    x, y, z = random_point_in_cylinder(center, radius, height, direction)
    assert x == pytest.approx(0.29184067570623795, abs=err)
    assert y == pytest.approx(-1.5489545079008842, abs=err)
    assert z == pytest.approx(0.1757693460699583, abs=err)

    direction = "y"
    x, y, z = random_point_in_cylinder(center, radius, height, direction)
    assert x == pytest.approx(-1.417921218944154, abs=err)
    assert y == pytest.approx(-0.8295636754899149, abs=err)
    assert z == pytest.approx(-1.2828806053300128, abs=err)

    direction = "x"
    x, y, z = random_point_in_cylinder(center, radius, height, direction)
    assert x == pytest.approx(0.5063181926884202, abs=err)
    assert y == pytest.approx(-0.5839696769917422, abs=err)
    assert z == pytest.approx(1.10139156645954, abs=err)


def test_validate_value_positive():
    """Test point in test value."""
    validate_value("test_value", 1.0)  # Should not raise an exception


def test_validate_value_negative():
    """Test point in test value."""
    with pytest.raises(Exception, match="Invalid test_value, needs to be positive"):
        validate_value("test_value", -1.0)


def test_load_molecule_from_file(tmp_path):
    """Test point in load molecule."""
    molecule = build_molecule("H2O")
    molecule_file = tmp_path / "water.xyz"
    write(molecule_file, molecule)
    loaded_molecule = load_molecule(str(molecule_file))
    assert len(loaded_molecule) == 3


def test_load_molecule_from_name():
    """Test point in load molecule."""
    molecule = load_molecule("H2O")
    assert len(molecule) == 3


def test_get_insertion_position_sphere():
    """Test point in sphere."""
    center = (5, 5, 5)
    radius = 10.0
    x, y, z = get_insertion_position("sphere", center, radius=radius)
    assert x == pytest.approx(-1.624620295372745, abs=err)
    assert y == pytest.approx(-1.3082748266588151, abs=err)
    assert z == pytest.approx(3.9856114049456455, abs=err)


def test_get_insertion_position_box():
    """Test point in box."""
    center = (5, 5, 5)
    a = 10.0
    x, y, z = get_insertion_position("box", center, a=a, b=a, c=a)
    assert x == pytest.approx(7.796391455704136, abs=err)
    assert y == pytest.approx(6.21107761885786, abs=err)
    assert z == pytest.approx(5.351538692139917, abs=err)


def test_get_insertion_position_ellipsoid():
    """Test point in ellipsoid."""
    center = (5, 5, 5)
    x, y, z = get_insertion_position("ellipsoid", center, a=2.0, b=2.0, c=4.0)
    assert x == pytest.approx(5.279146598518497, abs=err)
    assert y == pytest.approx(3.5184197470278056, abs=err)
    assert z == pytest.approx(3.7940043461327075, abs=err)


def test_get_insertion_position_cylinder():
    """Test point in cylinder."""
    center = (5, 5, 5)
    x, y, z = get_insertion_position("cylinderZ", center, radius=3.0, height=10.0)
    assert x == pytest.approx(5.437761013559357, abs=err)
    assert y == pytest.approx(2.6765682381486737, abs=err)
    assert z == pytest.approx(5.351538692139917, abs=err)


def test_set_defaults_centre():
    """Test centre."""
    cell = [10, 10, 10]
    centre, _, _, _, _, _ = set_defaults(cell, centre=None)
    assert centre[0] == pytest.approx(5.0, abs=err)
    assert centre[1] == pytest.approx(5.0, abs=err)
    assert centre[2] == pytest.approx(5.0, abs=err)


def test_set_defaults_box():
    """Test box."""
    cell = [10, 10, 10]
    centre, a, b, c, _, _ = set_defaults(cell, where="box", b=5.0, centre=None)
    assert centre[0] == pytest.approx(5.0, abs=err)
    assert a == pytest.approx(10.0, abs=err)
    assert b == pytest.approx(5.0, abs=err)
    assert c == pytest.approx(10.0, abs=err)


def test_set_defaults_anywhere():
    """Test box."""
    cell = [10, 10, 10]
    centre, a, b, c, _, _ = set_defaults(cell, where="anywhere", centre=None)
    assert centre[0] == pytest.approx(5.0, abs=err)
    assert a == pytest.approx(10.0, abs=err)
    assert b == pytest.approx(10.0, abs=err)
    assert c == pytest.approx(10.0, abs=err)


def test_set_defaults_ellipsoid():
    """Test ellipsoid."""
    cell = [10, 10, 10]
    centre, a, b, c, _, _ = set_defaults(
        cell, where="ellipsoid", b=3.0, a=3.0, centre=None
    )
    assert centre[0] == pytest.approx(5.0, abs=err)
    assert a == pytest.approx(3.0, abs=err)
    assert b == pytest.approx(3.0, abs=err)
    assert c == pytest.approx(5.0, abs=err)


def test_set_defaults_cylinder():
    """Test ellipsoid."""
    cell = [10, 12, 14]
    centre, _, _, _, radius, height = set_defaults(cell, where="cylinderZ", centre=None)
    assert centre[0] == pytest.approx(5.0, abs=err)
    assert radius == pytest.approx(5.0, abs=err)
    assert height == pytest.approx(7.0, abs=err)

    centre, _, _, _, radius, height = set_defaults(cell, where="cylinderX", centre=None)
    assert centre[0] == pytest.approx(5.0, abs=err)
    assert radius == pytest.approx(6.0, abs=err)
    assert height == pytest.approx(5.0, abs=err)

    centre, _, _, _, radius, height = set_defaults(cell, where="cylinderY", centre=None)
    assert centre[0] == pytest.approx(5.0, abs=err)
    assert radius == pytest.approx(5.0, abs=err)
    assert height == pytest.approx(6.0, abs=err)


def test_rotate_molecule():
    """Test rotate molecule."""
    molecule = build_molecule("H2O")
    a1 = molecule.get_angle(2, 0, 1)
    a2 = molecule.get_angle(2, 1, 0)

    rotated_molecule = rotate_molecule(molecule)
    assert a1 == pytest.approx(rotated_molecule.get_angle(2, 0, 1))
    assert a2 == pytest.approx(rotated_molecule.get_angle(2, 1, 0))


def test_optimize_geometry(tmp_path):
    """Test go."""
    molecule = build_molecule("H2O")
    molecule.set_cell([10, 10, 10])
    molecule.set_pbc([True, True, True])
    structure_file = tmp_path / "water.cif"
    write(structure_file, molecule)
    optimized_energy, _ = optimize_geometry(
        str(structure_file),
        device="cpu",
        arch="mace_mp",
        model="small-0b2",
        fmax=0.01,
        out_path=tmp_path,
        opt_cell=True,
    )
    assert optimized_energy == pytest.approx(-14.17098106193308, abs=err)


def test_pack_molecules(tmp_path):
    """Test pack molecule."""
    system = Atoms(
        "Ca", positions=[(5.0, 5.0, 5.0)], cell=[10, 10, 10], pbc=[True, True, True]
    )
    system_file = tmp_path / "system.cif"
    write(system_file, system)

    e, _ = pack_molecules(
        system=str(system_file),
        molecule="H2O",
        nmols=2,
        arch="mace_mp",
        model="small-0b2",
        device="cpu",
        where="sphere",
        center=(5.0, 5.0, 5.0),
        radius=5.0,
        seed=2042,
        temperature=300,
        ntries=10,
        geometry=False,
        fmax=0.1,
        out_path=tmp_path,
    )

    assert (tmp_path / "system+1H2O.cif").exists()
    assert (tmp_path / "system+2H2O.cif").exists()
    assert e == pytest.approx(-29.21589570470306, abs=err)


def test_pack_molecules_atoms(tmp_path):
    """Test pack molecule."""
    system = Atoms(
        "Ca", positions=[(5.0, 5.0, 5.0)], cell=[10, 10, 10], pbc=[True, True, True]
    )

    e, _ = pack_molecules(
        system=system,
        molecule="H2O",
        nmols=2,
        arch="mace_mp",
        model="small-0b2",
        device="cpu",
        where="sphere",
        center=(5.0, 5.0, 5.0),
        radius=5.0,
        seed=2042,
        temperature=300,
        ntries=10,
        geometry=False,
        fmax=0.1,
        out_path=tmp_path,
    )

    assert e == pytest.approx(-29.21589570470306, abs=err)


def test_pack_molecules_2(tmp_path, capsys):
    """Test pack molecule."""
    system = Atoms(
        "Ca", positions=[(2.5, 2.5, 2.5)], cell=[5, 5, 5], pbc=[True, True, True]
    )
    system_file = tmp_path / "system.cif"
    write(system_file, system)

    e, _ = pack_molecules(
        system=str(system_file),
        molecule="H2O",
        nmols=3,
        arch="mace_mp",
        model="small-0b2",
        device="cpu",
        where="sphere",
        center=(2.5, 2.5, 2.5),
        radius=2.5,
        seed=2042,
        temperature=300,
        ntries=2,
        geometry=False,
        fmax=0.1,
        out_path=tmp_path,
    )
    captured = capsys.readouterr()

    assert "Failed to insert particle 3 after 2 tries" in captured.out
    assert e == pytest.approx(-47.194755808249454, abs=err)


def test_save_the_day(tmp_path):
    """Test save the day."""
    molecule = build_molecule("H2O")
    molecule.set_cell([10, 10, 10])
    molecule.set_pbc([True, True, True])
    molecule.center()
    structure_file = tmp_path / "water.cif"
    write(structure_file, molecule)
    s = save_the_day(
        str(structure_file),
        device="cpu",
        arch="mace_mp",
        model="small-0b2",
        fmax=0.01,
        out_path=tmp_path,
    )
    assert s[0].position == pytest.approx([5.0, 4.99619815, 5.30704738], abs=err)


def test_save_the_day_md(tmp_path):
    """Test save the day."""
    molecule = build_molecule("H2O")
    molecule.set_cell([10, 10, 10])
    molecule.set_pbc([True, True, True])
    molecule.center()
    structure_file = tmp_path / "water.cif"
    write(structure_file, molecule)
    s = save_the_day(
        str(structure_file),
        device="cpu",
        arch="mace_mp",
        model="small-0b2",
        relax_strategy="md",
        md_timestep=1.0,
        md_steps=10.0,
        md_temperature=100.0,
    )
    assert s[0].position == pytest.approx([4.99684244, 5.00440785, 5.2987255], abs=err)


def test_save_the_day_invalid(tmp_path):
    """Test save the day."""
    molecule = build_molecule("H2O")
    molecule.set_cell([10, 10, 10])
    molecule.set_pbc([True, True, True])
    molecule.center()
    structure_file = tmp_path / "water.cif"
    write(structure_file, molecule)
    s = save_the_day(
        str(structure_file),
        device="cpu",
        arch="mace_mp",
        model="small-0b2",
        relax_strategy="invalid_stratregy",
        md_timestep=1.0,
        md_steps=10.0,
        md_temperature=100.0,
    )
    assert s is None


def test_run_md(tmp_path):
    """Test md."""
    molecule = build_molecule("H2O")
    molecule.set_cell([10, 10, 10])
    molecule.set_pbc([True, True, True])
    molecule.center()
    s = run_md_nve(
        molecule,
        device="cpu",
        arch="mace_mp",
        model="small-0b2",
        timestep=1.0,
        steps=10.0,
        temp=100.0,
    )
    assert s[0].position == pytest.approx([4.99684244, 5.00440785, 5.2987255], abs=err)



"""Test cli for packmm."""

# Author; alin m elena, alin@elena.re
# Contribs;
# Date: 22-02-2025
# ©alin m elena,
from __future__ import annotations

from typer.testing import CliRunner

from pack_mm.cli.packmm import app
from tests.utils import strip_ansi_codes

runner = CliRunner()


def test_packmm_default_values():
    """Check values."""
    result = runner.invoke(app)
    assert result.exit_code == 0
    assert "nothing to do" in strip_ansi_codes(result.output)


def test_packmm_custom_molecule():
    """Check molecule."""
    result = runner.invoke(app, ["--molecule", "CO2"])
    assert result.exit_code == 0
    assert "molecule='CO2'" in strip_ansi_codes(result.output)


def test_packmm_custom_nmols():
    """Check nmols."""
    result = runner.invoke(app, ["--nmols", "-2"])
    assert result.exit_code == 1
    assert "nmols=-2" in strip_ansi_codes(result.output)


def test_packmm_custom_ntries():
    """Check ntries."""
    result = runner.invoke(app, ["--ntries", "1"])
    assert result.exit_code == 0
    assert "ntries=1" in strip_ansi_codes(result.output)


def test_packmm_custom_seed():
    """Check seed."""
    result = runner.invoke(app, ["--seed", "1234"])
    assert result.exit_code == 0
    assert "seed=1234" in strip_ansi_codes(result.output)


def test_packmm_custom_every():
    """Check seed."""
    result = runner.invoke(app, ["--every", "10"])
    assert result.exit_code == 0
    assert "every=10" in strip_ansi_codes(result.output)


def test_packmm_custom_insertion_method():
    """Check insertion."""
    result = runner.invoke(app, ["--where", "sphere"])
    assert result.exit_code == 0
    assert "where=sphere" in strip_ansi_codes(result.output)


def test_packmm_custom_insert_strategy():
    """Check insertion."""
    result = runner.invoke(app, ["--insert-strategy", "hmc"])
    assert result.exit_code == 0
    assert "insert_strategy=hmc" in strip_ansi_codes(result.output)


def test_packmm_custom_relax_strategy():
    """Check relax."""
    result = runner.invoke(app, ["--relax-strategy", "md"])
    assert result.exit_code == 0
    assert "relax_strategy=md" in strip_ansi_codes(result.output)


def test_packmm_custom_center():
    """Check centre."""
    result = runner.invoke(app, ["--centre", "0.5,0.5,0.5"])
    assert result.exit_code == 0
    assert "centre='0.5,0.5,0.5'" in strip_ansi_codes(result.output)


def test_packmm_custom_radius():
    """Check radius."""
    result = runner.invoke(app, ["--radius", "10.0"])
    assert result.exit_code == 0
    assert "radius=10.0" in strip_ansi_codes(result.output)


def test_packmm_custom_height():
    """Check height."""
    result = runner.invoke(app, ["--height", "5.0"])
    assert result.exit_code == 0
    assert "height=5.0" in strip_ansi_codes(result.output)


def test_packmm_mlip():
    """Check mlip."""
    result = runner.invoke(
        app, ["--arch", "mace", "--model", "some", "--device", "cuda"]
    )
    assert result.exit_code == 0
    assert "arch='mace'" in strip_ansi_codes(result.output)
    assert "model='some'" in strip_ansi_codes(result.output)
    assert "device='cuda'" in strip_ansi_codes(result.output)


def test_packmm_out_path():
    """Check out_path."""
    result = runner.invoke(app, ["--out-path", "out"])
    assert result.exit_code == 0
    assert "out_path='out'" in strip_ansi_codes(result.output)


def test_packmm_custom_box_dimensions():
    """Check box."""
    result = runner.invoke(app, ["--a", "30.0", "--b", "30.0", "--c", "30.0"])
    assert result.exit_code == 0
    assert "a=30.0" in strip_ansi_codes(result.output)
    assert "b=30.0" in strip_ansi_codes(result.output)
    assert "c=30.0" in strip_ansi_codes(result.output)


def test_packmm_empty_box_dimensions():
    """Check box empty."""
    result = runner.invoke(
        app, ["--cell-a", "30.0", "--cell-b", "30.0", "--cell-c", "30.0"]
    )
    assert result.exit_code == 0
    assert "cell_a=30.0" in strip_ansi_codes(result.output)
    assert "cell_b=30.0" in strip_ansi_codes(result.output)
    assert "cell_c=30.0" in strip_ansi_codes(result.output)


def test_packmm_custom_temperature():
    """Check temperature."""
    result = runner.invoke(app, ["--temperature", "400.0"])
    assert result.exit_code == 0
    assert "temperature=400.0" in strip_ansi_codes(result.output)


def test_packmm_md_temperature():
    """Check md temperature."""
    result = runner.invoke(app, ["--md-temperature", "300.0"])
    assert result.exit_code == 0
    assert "md_temperature=300.0" in strip_ansi_codes(result.output)


def test_packmm_md_timestep():
    """Check md temperature."""
    result = runner.invoke(app, ["--md-timestep", "1.0"])
    assert result.exit_code == 0
    assert "md_timestep=1.0" in strip_ansi_codes(result.output)


def test_packmm_md_steps():
    """Check md steps."""
    result = runner.invoke(app, ["--md-steps", "10"])
    assert result.exit_code == 0
    assert "md_steps=10" in strip_ansi_codes(result.output)


def test_packmm_custom_fmax():
    """Check fmax."""
    result = runner.invoke(app, ["--fmax", "0.05"])
    assert result.exit_code == 0
    assert "fmax=0.05" in strip_ansi_codes(result.output)


def test_packmm_no_geometry_optimization():
    """Check optimisation."""
    result = runner.invoke(app, ["--no-geometry"])
    assert result.exit_code == 0
    assert "geometry=False" in strip_ansi_codes(result.output)


def test_packmm_invalid_insertion_method():
    """Check insertion."""
    result = runner.invoke(app, ["--where", "invalid_method"])
    assert result.exit_code != 0
    assert "Invalid value for '--where'" in strip_ansi_codes(result.output)


def test_packmm_invalid_insertion_strategy():
    """Check insertion strategt."""
    result = runner.invoke(app, ["--insert-strategy", "invalid_method"])
    assert result.exit_code != 0
    assert "Invalid value for '--insert-strategy'" in strip_ansi_codes(result.output)


def test_packmm_invalid_relax_strategy():
    """Check insertion strategt."""
    result = runner.invoke(app, ["--relax-strategy", "invalid_method"])
    assert result.exit_code != 0
    assert "Invalid value for '--relax-strategy'" in strip_ansi_codes(result.output)


def test_packmm_invalid_md_steps():
    """Check md steps."""
    result = runner.invoke(app, ["--nmols", "1", "--md-steps", "-10"])
    assert "Invalid MD steps" in strip_ansi_codes(result.output)


def test_packmm_invalid_md_temperature():
    """Check md steps."""
    result = runner.invoke(app, ["--nmols", "1", "--md-temperature", "-10.0"])
    assert "Invalid MD temperature" in strip_ansi_codes(result.output)


def test_packmm_invalid_md_timestep():
    """Check md timestep."""
    result = runner.invoke(app, ["--nmols", "1", "--md-timestep", "-10.0"])
    assert "Invalid MD timestep" in strip_ansi_codes(result.output)


def test_packmm_invalid_centre_format():
    """Check centre."""
    result = runner.invoke(app, ["--nmols", "1", "--centre", "0.5,0.5"])
    assert result.exit_code != 0
    assert "Invalid centre" in strip_ansi_codes(result.output)


def test_packmm_invalid_radius():
    """Check box radius."""
    result = runner.invoke(app, ["--nmols", "1", "--radius", "-10.0"])
    assert result.exit_code != 0
    assert "Invalid radius" in strip_ansi_codes(result.output)


def test_packmm_invalid_height():
    """Check box height."""
    result = runner.invoke(app, ["--nmols", "1", "--height", "-5.0"])
    assert result.exit_code != 0
    assert "Invalid height" in strip_ansi_codes(result.output)


def test_packmm_invalid_box_dimensions_a():
    """Check box dimension."""
    result = runner.invoke(app, ["--nmols", "1", "--a", "-30.0"])
    assert result.exit_code != 0
    assert "Invalid box a" in strip_ansi_codes(result.output)


def test_packmm_invalid_box_dimensions_b():
    """Check box dimension."""
    result = runner.invoke(app, ["--nmols", "1", "--b", "-30.0"])
    assert result.exit_code != 0
    assert "Invalid box b" in strip_ansi_codes(result.output)


def test_packmm_invalid_box_dimensions_c():
    """Check box dimension."""
    result = runner.invoke(app, ["--nmols", "1", "--c", "-30.0"])
    assert result.exit_code != 0
    assert "Invalid box c" in strip_ansi_codes(result.output)


def test_packmm_invalid_temperature():
    """Check temperature."""
    result = runner.invoke(app, ["--nmols", "1", "--temperature", "-400.0"])
    assert result.exit_code != 0
    assert "Invalid temperature" in strip_ansi_codes(result.output)


def test_packmm_invalid_fmax():
    """Check fmax."""
    result = runner.invoke(app, ["--nmols", "1", "--fmax", "-0.05"])
    assert result.exit_code != 0
    assert "Invalid fmax" in strip_ansi_codes(result.output)


def test_packmm_invalid_ntries():
    """Check ntries."""
    result = runner.invoke(app, ["--nmols", "1", "--ntries", "-1"])
    assert result.exit_code != 0
    assert "Invalid ntries" in strip_ansi_codes(result.output)



"""Command line for packmm."""

# Author; alin m elena, alin@elena.re
# Contribs;
# Date: 22-02-2025
# ©alin m elena,
from __future__ import annotations

from enum import Enum

import typer

from pack_mm.core.core import pack_molecules


class InsertionMethod(str, Enum):
    """Insertion options."""

    ANYWHERE = "anywhere"
    SPHERE = "sphere"
    BOX = "box"
    CYLINDER_Z = "cylinderZ"
    CYLINDER_Y = "cylinderY"
    CYLINDER_X = "cylinderX"
    ELLIPSOID = "ellipsoid"


class InsertionStrategy(str, Enum):
    """Insertion options."""

    # propose randomly a point
    MC = "mc"
    # hybrid monte carlo
    HMC = "hmc"


class RelaxStrategy(str, Enum):
    """Relaxation options."""

    GEOMETRY_OPTIMISATION = "geometry_optimisation"
    MD = "md"


app = typer.Typer(no_args_is_help=True)


@app.command()
def packmm(
    system: str | None = typer.Option(
        None,
        help="""The original box in which you want to add particles.
        If not provided, an empty box will be created.""",
    ),
    molecule: str = typer.Option(
        "H2O",
        help="""Name of the molecule to be processed, ASE-recognizable or
        ASE-readable file.""",
    ),
    nmols: int = typer.Option(-1, help="Target number of molecules to insert."),
    ntries: int = typer.Option(
        50, help="Maximum number of attempts to insert each molecule."
    ),
    every: int = typer.Option(
        -1, help="Run MD-NVE or Geometry optimisation everyth insertion."
    ),
    seed: int = typer.Option(2025, help="Random seed for reproducibility."),
    md_steps: int = typer.Option(10, help="Number of steps to run MD."),
    md_timestep: float = typer.Option(1.0, help="Timestep for MD integration, in fs."),
    where: InsertionMethod = typer.Option(
        InsertionMethod.ANYWHERE,
        help="""Where to insert the molecule. Choices: 'anywhere', 'sphere',
        'box', 'cylinderZ', 'cylinderY', 'cylinderX', 'ellipsoid'.""",
    ),
    insert_strategy: InsertionStrategy = typer.Option(
        InsertionStrategy.MC,
        help="""How to insert a new molecule. Choices: 'mc', 'hmc',""",
    ),
    relax_strategy: RelaxStrategy = typer.Option(
        RelaxStrategy.GEOMETRY_OPTIMISATION,
        help="""How to relax the system to get more favourable structures.
            Choices: 'geometry_optimisation', 'md',""",
    ),
    centre: str | None = typer.Option(
        None,
        help="""Centre of the insertion zone, coordinates in Å,
        e.g., '5.0, 5.0, 5.0'.""",
    ),
    radius: float | None = typer.Option(
        None,
        help="""Radius of the sphere or cylinder in Å,
        depending on the insertion volume.""",
    ),
    height: float | None = typer.Option(None, help="Height of the cylinder in Å."),
    a: float | None = typer.Option(
        None,
        help="""Side of the box or semi-axis of the ellipsoid, in Å,
        depends on the insertion method.""",
    ),
    b: float | None = typer.Option(
        None,
        help="""Side of the box or semi-axis of the ellipsoid, in Å,
        depends on the insertion method.""",
    ),
    c: float | None = typer.Option(
        None,
        help="""Side of the box or semi-axis of the ellipsoid, in Å,
        depends on the insertion method.""",
    ),
    device: str = typer.Option(
        "cpu", help="Device to run calculations on (e.g., 'cpu' or 'cuda')."
    ),
    model: str = typer.Option("medium-omat-0", help="ML model to use."),
    arch: str = typer.Option("mace_mp", help="MLIP architecture to use."),
    temperature: float = typer.Option(
        300.0, help="Temperature for the Monte Carlo acceptance rule."
    ),
    md_temperature: float = typer.Option(
        100.0, help="Temperature for the Molecular dynamics relaxation."
    ),
    cell_a: float = typer.Option(
        20.0, help="Side of the empty box along the x-axis in Å."
    ),
    cell_b: float = typer.Option(
        20.0, help="Side of the empty box along the y-axis in Å."
    ),
    cell_c: float = typer.Option(
        20.0, help="Side of the empty box along the z-axis in Å."
    ),
    fmax: float = typer.Option(
        0.1, help="force tollerance for optimisation if needed."
    ),
    geometry: bool = typer.Option(
        True, help="Perform geometry optimization at the end."
    ),
    out_path: str = typer.Option(".", help="path to save various outputs."),
):
    """Pack molecules into a system based on the specified parameters."""
    print("Script called with following input")
    print(f"{system=}")
    print(f"{nmols=}")
    print(f"{molecule=}")
    print(f"{ntries=}")
    print(f"{seed=}")
    print(f"where={where.value}")
    print(f"{centre=}")
    print(f"{radius=}")
    print(f"{height=}")
    print(f"{a=}")
    print(f"{b=}")
    print(f"{c=}")
    print(f"{cell_a=}")
    print(f"{cell_b=}")
    print(f"{cell_c=}")
    print(f"{arch=}")
    print(f"{model=}")
    print(f"{device=}")
    print(f"{temperature=}")
    print(f"{fmax=}")
    print(f"{geometry=}")
    print(f"{out_path=}")
    print(f"{every=}")
    print(f"insert_strategy={insert_strategy.value}")
    print(f"relax_strategy={relax_strategy.value}")
    print(f"{md_steps=}")
    print(f"{md_timestep=}")
    print(f"{md_temperature=}")
    if nmols == -1:
        print("nothing to do, no molecule to insert")
        raise typer.Exit(0)

    center = centre
    if centre:
        center = tuple(map(float, centre.split(",")))
        lc = [x < 0.0 for x in center]
        if len(center) != 3 or any(lc):
            err = "Invalid centre 3 coordinates expected!"
            print(f"{err}")
            raise Exception("Invalid centre 3 coordinates expected!")

    pack_molecules(
        system=system,
        molecule=molecule,
        nmols=nmols,
        arch=arch,
        model=model,
        device=device,
        where=where,
        center=center,
        radius=radius,
        height=height,
        a=a,
        b=b,
        c=c,
        seed=seed,
        temperature=temperature,
        ntries=ntries,
        fmax=fmax,
        geometry=geometry,
        cell_a=cell_a,
        cell_b=cell_b,
        cell_c=cell_c,
        out_path=out_path,
        every=every,
        relax_strategy=relax_strategy,
        insert_strategy=insert_strategy,
        md_steps=md_steps,
        md_timestep=md_timestep,
        md_temperature=md_temperature,
    )
